from .base_endpoint import BaseEndpoint


class DoNotDisturb(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_user_do_not_disturb(self, user_id: str):
        """Returns the specificied users DND and Ring Splash state.

        Args:
            user_id (str): Target user id to return data.

        Returns:
            Dict: States DND and Ring Splash status.
        """

        endpoint = "/users/do-not-disturb"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)

    # POST

    # PUT

    def put_user_do_not_disturb(
        self, user_id: str, dnd_active: bool = False, ring_splash_active: bool = False
    ):
        """Updates a user's DND and Ring Splash status.

        Args:
            user_id (str): Target user id of user you would like to update the state of.
            dnd_active (bool): True to enable DND and False to disable DND. Defaults to False.
            ring_splash_active (bool): True to enable Ring Splash and False to disable Ring Splash. Defaults to False.

        Returns:
            Dict: New DND and Ringsplash state of the user.
        """

        data = {
            "isActive": dnd_active,
            "ringSplash": ring_splash_active,
            "userId": user_id,
        }

        endpoint = "/users/do-not-disturb"

        return self._requester.put(endpoint, data)

    # DELETE



from .report_utils.graphviz_module import GraphvizModule
from ..utils.helpers import find_entity_with_number_type
from .report_utils.parsing import call_flow_module


from ..store import DataStore
from ..store import broadwork_entities as bre


def main(
    api,
    service_provider_id: str,
    group_id: str,
    number: str,
    number_type: str,
    broadworks_entity_type: str,
):
    logger = api.logger
    # Creates data store for use later
    data_store = DataStore()

    logger.info("Fetching Service Provider & Group details")
    # Gather entities
    service_provider = bre.ServiceProvider.from_dict(
        data=api.service_providers.get_service_provider(service_provider_id)
    )
    group = bre.Group.from_dict(
        service_provider=service_provider,
        data=api.groups.get_group(service_provider_id, group_id),
    )

    data_store.store_objects(service_provider, group)

    logger.info("Fetching group auto attendants")
    auto_attendants = api.auto_attendants.get_auto_attendants(
        service_provider_id, group_id
    )
    for aa in auto_attendants:
        auto_attendant = bre.AutoAttendant.from_dict(
            group=group,
            data=api.auto_attendants.get_auto_attendant(aa["serviceUserId"]),
        )
        data_store.auto_attendants.append(auto_attendant)

    logger.info("Fetching all users this may take a couple of minutes")
    users = api.users.get_users(service_provider_id, group_id, extended=True)

    # Captures users with the forward fucntionality
    logger.info("Fetching call forward always users")
    call_forward_always_users = [
        item["user"]["userId"]
        for item in api.call_forwarding_always.get_bulk_call_forwarding_always(
            service_provider_id, group_id
        )
        if item["service"]["assigned"] and item["data"]["isActive"]
    ]

    logger.info("Fetching call forward busy users")
    call_forward_busy_users = [
        item["user"]["userId"]
        for item in api.call_forwarding_busy.get_bulk_call_forwarding_busy(
            service_provider_id, group_id
        )
        if item["service"]["assigned"] and item["data"]["isActive"]
    ]

    logger.info("Fetching call forward no answer users")
    call_forward_no_answer_users = [
        item["user"]["userId"]
        for item in api.call_forwarding_no_answer.get_bulk_call_forwarding_no_answer(
            service_provider_id, group_id
        )
        if item["service"]["assigned"] and item["data"]["isActive"]
    ]

    logger.info("Fetching call forward not reachable users")
    call_forward_not_reachable = [
        item["user"]["userId"]
        for item in api.call_forwarding_not_reachable.get_bulk_call_forwarding_not_reachable(
            service_provider_id, group_id
        )
        if item["service"]["assigned"] and item["data"]["isActive"]
    ]

    logger.info("Analysing all users pulling call forwarding numbers")
    for u in users:
        user = bre.User.from_dict(group=group, data=u)

        if user.id in call_forward_always_users:
            user.call_forwarding_always = str(
                api.call_forwarding_always.get_user_call_forwarding_always(user.id)[
                    "forwardToPhoneNumber"
                ]
            )
        if user.id in call_forward_busy_users:
            user.call_forwarding_busy = str(
                api.call_forwarding_busy.get_user_call_forwarding_busy(user.id)[
                    "forwardToPhoneNumber"
                ]
            )
        if user.id in call_forward_no_answer_users:
            user.call_forwarding_no_answer = str(
                api.call_forwarding_no_answer.get_user_call_forwarding_no_answer(
                    user.id
                )["forwardToPhoneNumber"]
            )
        if user.id in call_forward_not_reachable:
            user.call_forwarding_not_reachable = str(
                api.call_forwarding_not_reachable.get_user_call_forwarding_not_reachable(
                    user.id
                )["forwardToPhoneNumber"]
            )

        data_store.users.append(user)

    logger.info("Fetching call centers")
    call_centers = api.call_centers.get_group_call_centers(
        service_provider_id, group_id
    )

    logger.info("Analysing call centers")
    for cc in call_centers:
        call_center = api.call_centers.get_group_call_center(cc["serviceUserId"])
        call_center["agents"] = api.call_centers.get_group_call_center_agents(
            cc["serviceUserId"]
        )["agents"]

        call_center = bre.CallCenter.from_dict(group=group, data=call_center)

        try:
            overflow_settings = api.call_centers.get_group_call_center_overflow(
                call_center.service_user_id
            )
            call_center.overflow_calls_action = overflow_settings["action"]
            call_center.overflow_calls_transfer_to_phone_number = (
                overflow_settings["transferPhoneNumber"]
                if call_center.overflow_calls_action == "Transfer"
                else None
            )
            logger.info(f"Fetched call center {cc['serviceUserId']} overflow details")
        except Exception:
            logger.error(f"Call center {cc['serviceUserId']} has no overflow")
            call_center.overflow_calls_action = None
            call_center.overflow_calls_transfer_to_phone_number = None

        try:
            stranded_calls_settings = (
                api.call_centers.get_group_call_center_stranded_calls(
                    call_center.service_user_id
                )
            )
            call_center.stranded_calls_action = stranded_calls_settings["action"]
            call_center.stranded_calls_transfer_to_phone_number = (
                stranded_calls_settings["transferPhoneNumber"]
                if call_center.stranded_calls_action == "Transfer"
                else None
            )
            logger.info(
                f"Fetched call center {cc['serviceUserId']} stranded call details"
            )
        except Exception:
            logger.error(
                f"Call center {cc['serviceUserId']} has no stranded calls action"
            )
            call_center.stranded_calls_action = None
            call_center.stranded_calls_transfer_to_phone_number = None

        try:
            stranded_calls_unavailable_settings = (
                api.call_centers.get_group_call_center_stranded_calls_unavailable(
                    call_center.service_user_id
                )
            )
            action_value = stranded_calls_unavailable_settings["action"]
            call_center.stranded_call_unavailable_action = (
                None if action_value == "None" else action_value
            )
            call_center.stranded_call_unavailable_transfer_to_phone_number = (
                stranded_calls_unavailable_settings["transferPhoneNumber"]
                if call_center.stranded_call_unavailable_action == "Transfer"
                else None
            )
            logger.info(
                f"Fetched call center {cc['serviceUserId']} stranded calls unavailable details"
            )
        except Exception:
            logger.error(
                f"Call center {cc['serviceUserId']} has no stranded calls unavailable action"
            )
            call_center.stranded_call_unavailable_action = None
            call_center.stranded_call_unavailable_transfer_to_phone_number = None

        try:
            forced_forwarding_settings = (
                api.call_centers.get_group_call_center_forced_forwarding(
                    call_center.service_user_id
                )
            )
            call_center.forced_forwarding_enabled = forced_forwarding_settings[
                "enabled"
            ]
            call_center.forced_forwarding_forward_to_phone_number = (
                forced_forwarding_settings["forwardToPhoneNumber"]
                if call_center.forced_forwarding_enabled
                else None
            )
            logger.info(
                f"Fetched call center {cc['serviceUserId']} forced forwarding details"
            )
        except Exception:
            logger.error(
                f"Call center {cc['serviceUserId']} has no forced forwarding action"
            )
            call_center.forced_forwarding_enabled = False
            call_center.forced_forwarding_forward_to_phone_number = None

        data_store.call_centers.append(call_center)

    logger.info("Fetching hunt groups")
    hunt_groups = api.hunt_groups.get_group_hunt_groups(service_provider_id, group_id)
    for hg in hunt_groups:
        hunt_group = bre.HuntGroup.from_dict(
            group=group, data=api.hunt_groups.get_group_hunt_group(hg["serviceUserId"])
        )
        data_store.hunt_groups.append(hunt_group)

    # locate number using broadworks_entity_type to zone in on correct location
    call_flow_start_node = find_entity_with_number_type(
        number, number_type.lower(), getattr(data_store, broadworks_entity_type + "s")
    )
    call_flow_start_node._start_node = True
    logger.info("Call flow start node found")

    # Nodes used in the graph
    logger.info("Parsing nodes for call flow")
    bre_nodes = call_flow_module(call_flow_start_node, data_store)

    logger.info("Generating report")
    # build, generate, save graph
    graph = GraphvizModule("./os_reports/")
    graph.generate_call_flow_graph(bre_nodes, number)
    graph._save_graph(f"Calls To {number}")
    logger.info("Report saved")

    return True



from ..exceptions import OSServiceNotAssigned


def main(api: object, service_provider_id: str) -> dict:
    """Returns a breakdown of the Trunking Call Capacity of a Service Provider/ Enterprise (SP/ENT).
    This will show the totals at each level from SP/ ENT to Group to Trunk Groups located in Groups.
    At each level Max Active Calls and Bursting Max Active calls are detailed and then differences at
    calculated.
    """

    logger = api.logger

    return_data = {}

    logger.info(f"Fetching {service_provider_id} call capacity")

    # OSServiceNotAssigned is returned if the Trunk Group service is not assigend to SP/ ENT
    try:
        service_provider_capacity = (
            api.trunk_groups.get_service_provider_trunk_group_call_capacity(
                service_provider_id
            )
        )
    except Exception:
        logger.error(f"{OSServiceNotAssigned}")
        raise OSServiceNotAssigned

    return_data["serviceProviderId"] = service_provider_capacity["serviceProviderId"]
    return_data["maxActiveCalls"] = service_provider_capacity["maxActiveCalls"]
    return_data["burstingMaxActiveCalls"] = service_provider_capacity[
        "burstingMaxActiveCalls"
    ]
    return_data["groupsCallCapacityTotal"] = 0
    return_data["groupsBurstingCallCapacityTotal"] = 0
    return_data["groups"] = []

    logger.info(f"Fetching complete list of groups in {service_provider_id}")
    groups_in_service_provider = api.groups.get_groups(service_provider_id)

    # getting groups and group call capacities
    logger.info("Fetching groups trunking capacities")
    for group in groups_in_service_provider:
        formatted_group = {}
        formatted_group["groupId"] = group["groupId"]
        formatted_group["groupName"] = group["groupName"]

        try:
            logger.info(f"Fetching group {group} trunking capacity")
            group_capacity = api.trunk_groups.get_group_trunk_groups_call_capacity(
                service_provider_id, group["groupId"]
            )
            formatted_group["maxActiveCalls"] = group_capacity["maxActiveCalls"]
            formatted_group["burstingMaxAvailableActiveCalls"] = group_capacity[
                "burstingMaxAvailableActiveCalls"
            ]
            formatted_group["burstingMaxActiveCalls"] = group_capacity[
                "burstingMaxActiveCalls"
            ]
            formatted_group["burstingMaxAvailableActiveCalls"] = group_capacity[
                "burstingMaxAvailableActiveCalls"
            ]

            return_data["groupsCallCapacityTotal"] += group_capacity["maxActiveCalls"]
            return_data["groupsBurstingCallCapacityTotal"] += group_capacity[
                "burstingMaxAvailableActiveCalls"
            ]

        # group has no trunks and therefore fails
        except Exception:
            logger.error(f"Group {group} has no trunks")
            formatted_group["maxActiveCalls"] = 0
            formatted_group["burstingMaxAvailableActiveCalls"] = 0
            formatted_group["burstingMaxActiveCalls"] = 0
            formatted_group["burstingMaxAvailableActiveCalls"] = 0

            return_data["groupsCallCapacityTotal"] += 0
            return_data["groupsBurstingCallCapacityTotal"] += 0

        formatted_group["trunkGroups"] = []
        return_data["groups"].append(formatted_group)

    # getting trunk group capacities
    logger.info("Fetching trunk groups call capacities")
    for group in return_data["groups"]:
        group["trunkGroupsCallCapacityTotal"] = 0
        group["trunkGroupsBurstingCallCapacityTotal"] = 0

        try:
            group_trunk_groups = api.trunk_groups.get_group_trunk_groups(
                service_provider_id, group["groupId"]
            )
            for trunk_group in group_trunk_groups:
                trunk_group_detailed = api.trunk_groups.get_group_trunk_group(
                    service_provider_id, group["groupId"], trunk_group["name"]
                )

                group["trunkGroupsCallCapacityTotal"] += trunk_group_detailed[
                    "maxActiveCalls"
                ]

                group["trunkGroups"].append(
                    {
                        "name": trunk_group_detailed["name"],
                        "maxActiveCalls": trunk_group_detailed["maxActiveCalls"],
                        "burstingMaxActiveCalls": trunk_group_detailed[
                            "burstingMaxActiveCalls"
                        ]
                        if "burstingMaxActiveCalls" in trunk_group_detailed
                        else 0,
                    }
                )

            group["callCapacityDifference"] = (
                group["maxActiveCalls"] - group["trunkGroupsCallCapacityTotal"]
            )
            group["burstingCallCapacityDifference"] = (
                group["trunkGroupsBurstingCallCapacityTotal"]
                - group["trunkGroupsBurstingCallCapacityTotal"]
            )

        # if no trunks exist in group api call fails
        except Exception:
            logger.error(f"No trunk groups in group {group} found")
            group["trunkGroupsCallCapacityTotal"] = 0
            group["trunkGroupsBurstingCallCapacityTotal"] = 0
            group["callCapacityDifference"] = 0
            group["burstingCallCapacityDifference"] = 0

    # working out totals difference
    return_data["callCapacityDifference"] = (
        return_data["maxActiveCalls"] - return_data["groupsCallCapacityTotal"]
    )
    return_data["burstingCallCapacityDifference"] = (
        return_data["burstingMaxActiveCalls"]
        - return_data["groupsBurstingCallCapacityTotal"]
    )

    return return_data



supported_filters = [
    "macAddress",
    "lastName",
    "firstName",
    "dn",
    "emailAddress",
    "userId",
    "extension",
]

supported_filter_types = [
    "contains",
    "startsWith",
    "endsWith",
    "equals",
]

group_services = [
    "Account/Authorization Codes",
    "Auto Attendant - Basic",
    "Auto Attendant - Standard",
    "Call Park",
    "Call Pickup",
    "Emergency Zones",
    "Group Paging",
    "Hunt Group",
    "Incoming Calling Plan",
    "Meet-Me Conferencing",
    "Music On Hold",
    "Outgoing Calling Plan",
    "Trunk Group",
    "Voice Messaging Group",
    "VoiceXML",
]

user_services = [
    "Advice Of Charge",
    "Alternate Numbers",
    "Anonymous Call Rejection",
    "Authentication",
    "Automatic Callback",
    "Automatic Hold/Retrieve",
    "Barge-in Exempt",
    "Basic Call Logs",
    "BroadWorks Agent",
    "BroadWorks Anywhere",
    "BroadWorks Mobility",
    "BroadWorks Receptionist - Enterprise",
    "BroadWorks Receptionist - Office",
    "BroadWorks Receptionist - Small Business",
    "BroadWorks Supervisor",
    "Busy Lamp Field",
    "Call Center - Basic",
    "Call Center Monitoring",
    "Call Center - Premium",
    "Call Center - Standard",
    "Call Forwarding Always",
    "Call Forwarding Always Secondary",
    "Call Forwarding Busy",
    "Call Forwarding No Answer",
    "Call Forwarding Not Reachable",
    "Call Forwarding Selective",
    "Calling Line ID Blocking Override",
    "Calling Line ID Delivery Blocking",
    "Calling Name Delivery",
    "Calling Name Retrieval",
    "Calling Number Delivery",
    "Calling Party Category",
    "Call Me Now",
    "Call Notify",
    "Call Recording",
    "Call Return",
    "Call Transfer",
    "Call Waiting",
    "Charge Number",
    "Classmark",
    "Client License 3",
    "CommPilot Express",
    "Communication Barring User-Control",
    "Connected Line Identification Presentation",
    "Connected Line Identification Restriction",
    "Customer Originated Trace",
    "Custom Ringback User",
    "Custom Ringback User - Call Waiting",
    "Custom Ringback User - Video",
    "Directed Call Pickup",
    "Directed Call Pickup with Barge-in",
    "Directory Number Hunting",
    "Direct Route",
    "Diversion Inhibitor",
    "Do Not Disturb",
    "Enhanced Call Logs",
    "Executive",
    "Executive-Assistant",
    "External Calling Line ID Delivery",
    "External Custom Ringback",
    "Fax Messaging",
    "Flash Call Hold",
    "Flexible Seating Guest",
    "Group Night Forwarding",
    "Hoteling Guest",
    "Hoteling Host",
    "In-Call Service Activation",
    "Integrated IM&P",
    "Intercept User",
    "Internal Calling Line ID Delivery",
    "Last Number Redial",
    "Location-Based Calling Restrictions",
    "Malicious Call Trace",
    "MobileLink",
    "Multiple Call Arrangement",
    "Music On Hold User",
    "MWI Delivery to Mobile Endpoint",
    "Number Portability Announcement",
    "N-Way Call",
    "Personal Assistant",
    "Physical Location",
    "Polycom Phone Services",
    "Pre-alerting Announcement",
    "Preferred Carrier User",
    "Prepaid",
    "Priority Alert",
    "Privacy",
    "Push to Talk",
    "Remote Office",
    "Resource Priority",
    "Route List",
    "Security Classification",
    "Selective Call Acceptance",
    "Selective Call Rejection",
    "Sequential Ring",
    "Shared Call Appearance",
    "Shared Call Appearance 10",
    "Shared Call Appearance 15",
    "Shared Call Appearance 20",
    "Shared Call Appearance 25",
    "Shared Call Appearance 30",
    "Shared Call Appearance 35",
    "Shared Call Appearance 5",
    "Silent Alerting",
    "Simultaneous Ring Personal",
    "SMDI Message Desk",
    "Speed Dial 100",
    "Speed Dial 8",
    "Third-Party MWI Control",
    "Third-Party Voice Mail Support",
    "Three-Way Call",
    "Two-Stage Dialing",
    "Video Add-On",
    "Video On Hold User",
    "Virtual On-Net Enterprise Extensions",
    "Voice Messaging User",
    "Voice Messaging User - Video",
    "Voice Portal Calling",
    "Zone Calling Restrictions",
]



def main(
    api,
    current_service_provider_id: str,
    current_group_id: str,
    target_service_provider_id: str,
    target_group_id: str,
    start_of_range_number: str,
    end_of_range_number: str = None,
) -> bool:
    """Moves singular or a range of numbers from one group to another located on the same Broadwork instance."""

    logger = api.logger

    logger.info(f"Removing numbers from {current_group_id}")
    # delete number from group
    api.dns.delete_group_dns(
        current_service_provider_id,
        current_group_id,
        start_of_range_number=start_of_range_number,
        end_of_range_number=end_of_range_number,
    )

    # check if sp/ent are the same if not number needs to be removed to sys level
    if not current_service_provider_id == target_service_provider_id:
        logger.info(f"Removing numbers from {current_service_provider_id}")
        # remove from sp/ent
        api.dns.delete_service_provider_dns(
            current_service_provider_id,
            start_of_range_number=start_of_range_number,
            end_of_range_number=end_of_range_number,
        )

        logger.info(
            f"Adding number to SP/ Ent: {target_service_provider_id} Group: {target_group_id}"
        )
        # assign to new group
        api.dns.post_group_dns_assign_bulk(
            target_service_provider_id,
            target_group_id,
            start_of_range_number=start_of_range_number,
            end_of_range_number=end_of_range_number,
        )
    else:
        # only when group is in the same sp/ent

        logger.info(f"Adding number to {target_group_id}")
        # assign to new group
        api.dns.post_group_dns(
            target_service_provider_id,
            target_group_id,
            start_of_range_number=start_of_range_number,
            end_of_range_number=end_of_range_number,
        )

    return True



from typing import Optional

from . import reports
from .api import API


class Reporter:
    """generates human friendly reports."""

    __instance = None

    @staticmethod
    def get_instance(api: API):
        """Static method to fetch the current instance."""
        if Reporter.__instance is None:
            Reporter(api)
        return Reporter.__instance

    def __init__(self, api: API) -> None:
        if Reporter.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            self.api = api
            Reporter.__instance = self

    def _run_report(self, report_name: str, *args, **kwargs):
        """Run a report function from the reports module."""
        self.api.logger.debug(
            f"Report {report_name} executed, args: {[args]}, kwargs: {kwargs}"
        )
        try:
            report_function = getattr(reports, report_name)
            self.api.logger.info(f"Report {report_name} executed")
        except AttributeError:
            self.api.logger.error(
                f"Report failed to execute, error: Report '{report_name}' not found."
            )
            raise AttributeError(
                f"Report '{report_name}' not found in 'reports' module."
            )
        return report_function(self.api, *args, **kwargs)

    def call_flow(
        self,
        *,
        service_provider_id: str,
        group_id: str,
        number: str,
        number_type: str,
        broadworks_entity_type: str,
    ) -> bool:
        """Generates a graphical flow chart of how a call to a specified number routes 
        through the broadworks system.
        
        NOTE: Report will be saved in .os_reports/ folder as a .svg file.

        Args:
            service_provider_id (str): Service Provider/ Enterprise where group is hosted.
            group_id (str): Group ID where target number for call flow is located. 
            number (str): Target number for call flow. 
            number_type (str): Type of number, options: "dn": Direct Number, "extension": Extension, "alias": Alias
            broadworks_entity_type (str): Broadworks entity type target number is associated with. \
            Options: "auto_attendant": Auto Attendant, "call_center": Call Center, "hunt_group": Hunt Group, "user": User
            
        Returns: Boolean True if report was generated successfully.
        """

        return self._run_report(
            "call_flow",
            service_provider_id,
            group_id,
            number,
            number_type,
            broadworks_entity_type,
        )

    def group_users_call_statistics(
        self,
        *,
        service_provider_id: str,
        group_id: str,
        start_date: str,
        end_date: Optional[str] = None,
        start_time: str = "00:00:00",
        end_time: str = "23:59:59",
        time_zone: str = "Z",
    ) -> bool:
        """Generates a CSV deatiling each users incoming and outgoing call statistics over 
        a specified period for a single group. Each row contains user extension, user ID, and call stats.
        
        NOTE: Report will be saved in .os_reports/ folder as a .csv file.

        Args:
            service_provider_id (str):  Service Provider/ Enterprise where group is hosted.
            group_id (str): Target Group you would like to know user statistics for.
            start_date (str): Start date of desired time period. Date must follow format 'YYYY-MM-DD'
            end_date (str, optional): End date of desired time period. Date must follow format 'YYYY-MM-DD'.\
                If this date is the same as Start date you do not need this parameter. Defaults to None.
            start_time (_type_, optional): Start time of desired time period. Time must follow formate 'HH:MM:SS'. \
                If you do not need to filter by time and want the whole day leave this parameter. Defaults to "00:00:00". MAX Request is 3 months.
            end_time (_type_, optional): End time of desired time period. Time must follow formate 'HH:MM:SS'. \
                If you do not need to filter by time and want the whole day leave this parameter. Defaults to "23:59:59". MAX Request is 3 months.
            time_zone (str, optional): A specified time you would like to see call records in. \
                Time zone must follow format 'GMT', 'EST', 'PST'. Defaults to "Z" (UTC Time Zone).
                
        Returns: Boolean True if report was generated successfully.
        """
        return self._run_report(
            "group_users_call_statistics",
            service_provider_id,
            group_id,
            start_date,
            end_date,
            start_time,
            end_time,
            time_zone,
        )

    def user_registration_report(
        self, *, service_provider_id: str, group_id: str
    ) -> bool:
        """Generates an Excel Worksheet detailing each Users ID, device name and registration status within a group.

        NOTE: Xlsx File into .os_reports/ named "Registration_report_for_(GroupID)"

        Args:
            service_provider_id (str):  Service Provider/ Enterprise where group is hosted.
            group_id (str): Target Group you would like to check the registration of.

        Returns: Boolean True if report was generated successfully.
        """
        return self._run_report(
            "user_registration_report", service_provider_id, group_id
        )



from .base_endpoint import BaseEndpoint


class CallForwardingSelective(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_user_call_forwarding_selective(self, user_id: str):
        """Retrieves the Forwarding Selective status for a specified User, alongside the criteria.

        Args:
            user_id (str): Target User ID

        Returns:
            Dict: Forwarding enabled status and the Forwarding criteria.
        """

        endpoint = "/users/call-forwarding-selective"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)

    def get_user_call_forwarding_selective_criterias(self, user_id: str):
        """Retrieves the Forwarding Selective status for a specified User, alongside the criteria's assigned.

        Args:
            user_id (str): Target User ID

        Returns:
            Dict: Forwarding enabled status and the Forwarding criteria's names and settings.
        """

        endpoint = "/users/call-forwarding-selective/criteria"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)

    def get_user_call_forwarding_selective_criteria(
        self, user_id: str, criteria_name: str
    ):
        """Retrieves the Forwarding Selective status for a specified User, alongside the specified Criteria

        Args:
            user_id (str): Target User ID
            criteria_name (str): Target Criteria Name

        Returns:
            Dict: Forwarding enabled status and the specified Criterias Settings.
        """

        endpoint = "/users/call-forwarding-selective/criteria"

        params = {"criteriaName": criteria_name, "userId": user_id}

        return self._requester.get(endpoint, params=params)


# POST

# PUT

# DELETE



import csv
import os


from ..utils.filers import copy_single_file_to_target_directory
from .report_utils.report_entities import call_records_statistics


def main(
    api: object,
    service_provider_id: str,
    group_id: str,
    start_date: str,
    end_date: str = None,
    start_time: str = "00:00:00",
    end_time: str = "23:59:59",
    time_zone: str = "Z",
):
    """Generates a CSV deatiling each users incoming and outgoing call statistics over 
        a specified period for a single group. Each row contains user extension, user ID, and call stats.

    Args:
        service_provider_id (str):  Service Provider/ Enterprise where group is hosted.
        group_id (str): Target Group you would like to know user statistics for.
        start_date (str): Start date of desired time period. Date must follow format 'YYYY-MM-DD'
        end_date (str, optional): End date of desired time period. Date must follow format 'YYYY-MM-DD'.\
            If this date is the same as Start date you do not need this parameter. Defaults to None.
        start_time (_type_, optional): Start time of desired time period. Time must follow formate 'HH:MM:SS'. \
            If you do not need to filter by time and want the whole day leave this parameter. Defaults to "00:00:00". MAX Request is 3 months.
        end_time (_type_, optional): End time of desired time period. Time must follow formate 'HH:MM:SS'. \
            If you do not need to filter by time and want the whole day leave this parameter. Defaults to "23:59:59". MAX Request is 3 months.
        time_zone (str, optional): A specified time you would like to see call records in. \
    """

    logger = api.logger

    # List of report_entities.call_records_statistics
    group_users_statistics = []

    logger.info(f"Fetching list of users in {group_id}")

    # Fetches complete list of users in group
    logger.info("fetching groups users")
    users = api.users.get_users(service_provider_id, group_id)
    failed_users = []

    # Pulls stats for each user, instantiates call_records_statistics, and append to group_users_statistics
    logger.info("Fetching users call statistics")
    for user in users:
        try:
            user_statistics = api.call_records.get_users_stats(
                user["userId"], start_date, end_date, start_time, end_time, time_zone
            )

            user_services = api.services.get_user_services(user_id=user["userId"])

        except Exception:
            logger.error(f"Failed to fetch {user} statistics - attempt 1/2")
            # attempt 2 in case of connection time out
            try:
                user_statistics = api.call_records.get_users_stats(
                    user["userId"],
                    start_date,
                    end_date,
                    start_time,
                    end_time,
                    time_zone,
                )

                user_services = api.services.get_user_services(user_id=user["userId"])

            except Exception:
                logger.error(f"Failed to fetch {user} statistics - attempt 2/2")
                failed_users.append(user)
                continue

        # Correction for API removing userId if no calls made by user
        if not user_statistics:
            user_statistics = {
                "userId": user["userId"],
                "total": 0,
                "totalAnsweredAndMissed": 0,
                "answeredTotal": 0,
                "missedTotal": 0,
                "busyTotal": 0,
                "redirectTotal": 0,
                "receivedTotal": 0,
                "receivedMissed": 0,
                "receivedAnswered": 0,
                "placedTotal": 0,
                "placedMissed": 0,
                "placedAnswered": 0,
            }

        user_statistics["servicePackServices"] = [
            service["serviceName"]
            for service in user_services["servicePackServices"]
            if service["assigned"]
        ]

        user_statistic_record = call_records_statistics.from_dict(
            user["firstName"], user["lastName"], user["extension"], user_statistics
        )
        group_users_statistics.append(user_statistic_record)

    output_directory = "./os_reports"
    file_name = os.path.join(
        output_directory,
        f"{group_id} User Call Statistics - {start_date} to {end_date}.csv",
    )
    logger.debug(f"Filename {file_name}")

    # Ensure the directory exists
    os.makedirs(output_directory, exist_ok=True)
    logger.debug("Directory exists")

    # Write statistics to csv
    logger.info("Writing report")
    with open(file_name, mode="w", newline="") as file:
        fieldnames = [
            field.name
            for field in call_records_statistics.__dataclass_fields__.values()
        ]
        writer = csv.DictWriter(file, fieldnames=fieldnames)
        writer.writeheader()

        for user in group_users_statistics:
            writer.writerow(user.__dict__)

        # Adds list of failed users to the bottom
        if failed_users:
            writer.writerow({})
            writer.writerow({fieldnames[1]: "Failed Users"})

            for failed_user in failed_users:
                writer.writerow({fieldnames[1]: failed_user["userId"]})

    # Add made_with_os.png for output
    copy_single_file_to_target_directory(
        "./odins_spear/assets/images/", "./os_reports/", "made_with_os.png"
    )
    logger.info(f"Report {file_name} saved to /os_reports")

    return True



from .base_endpoint import BaseEndpoint


class CallPickup(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_call_pickup_group_user(
        self, service_provider_id: str, group_id: str, user_id: str
    ):
        """Retrieves Pickup Group information for the specified user.

        Args:
            service_provider_id (str): Target Service Provider ID
            group_id (str): The Target Group ID the user is apart of.
            user_id (str): Target User ID

        Returns:
            Dict: Specified users pickup group, and the users within that group.
        """

        endpoint = "/groups/call-pickup/user"

        params = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "userId": user_id,
        }

        return self._requester.get(endpoint, params=params)


# POST

# PUT

# DELETE



import logging
from typing import Optional

from .requester import Requester
from .endpoints import *  # noqa: F403

from .exceptions import (
    OSApiAuthenticationFail,
    OSSessionRefreshFail,
    OSFailedToLocateSession,
)


class API:
    def __init__(
        self,
        base_url: str,
        username: str,
        password: str,
        rate_limit: bool = True,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        """ Connection to Odin API, all interactions with the api are here.

        Args:
            base_url (str): Base url of your odin instance api.
            username (str): Username used when logging into odin account.
            password (str): Password used when logging into odin account stored as virtual environment.
            rate_limit (bool): Enables (True) or Disables (False) rate limiting to 5 calls per second. Defaults to True.
            logger (logger, optional): Pass in external logger, if not default logger will be assigned. 
            
        Vars: 
            authorised (bool): Boolean value to indicate if api is authorised.\
            token (str): Token string returned from odin api.
        """

        self.base_url = base_url
        self.username = username
        self._password = password
        self.rate_limit = rate_limit
        self.authorised = False

        self.logger = logger if logger else self._setup_logger()
        self.logger.info(
            f"API initialised, user: {self.username}, base_url: {self.base_url}, rate_limit: {self.rate_limit}"
        )

        self._requester = Requester.get_instance(
            self.base_url, self.rate_limit, self.logger
        )

        # endpoints
        self.administrators = Administrators()  # noqa: F405
        self.alternate_numbers = AlternateNumbers()  # noqa: F405
        self.announcements = Announcements()  # noqa: F405
        self.authentication = Authentication()  # noqa: F405
        self.auto_attendants = AutoAttendants()  # noqa: F405
        self.call_centers = CallCenters()  # noqa: F405
        self.call_forwarding_always = CallForwardingAlways()  # noqa: F405
        self.call_forwarding_busy = CallForwardingBusy()  # noqa: F405
        self.call_forwarding_no_answer = CallForwardingNoAnswer()  # noqa: F405
        self.call_forwarding_not_reachable = CallForwardingNotReachable()  # noqa: F405
        self.call_forwarding_selective = CallForwardingSelective()  # noqa: F405
        self.call_pickup = CallPickup()  # noqa: F405
        self.call_processing_policies = CallProcessingPolicies()  # noqa: F405
        self.call_records = CallRecords()  # noqa: F405
        self.devices = Devices()  # noqa: F405
        self.dns = DNs()  # noqa: F405
        self.groups = Groups()  # noqa: F405
        self.emergency_zones = EmergencyZones()  # noqa: F405
        self.do_not_disturb = DoNotDisturb()  # noqa: F405
        self.hunt_groups = HuntGroups()  # noqa: F405
        self.service_providers = ServiceProviders()  # noqa: F405
        self.services = Services()  # noqa: F405
        self.session = Session()  # noqa: F405
        self.shared_call_appearance = SharedCallAppearance()  # noqa: F405
        self.schedules = Schedules()  # noqa: F405
        self.reports = Reports()  # noqa: F405
        self.regsitration = Registration()  # noqa: F405
        self.password_generate = PasswordGenerate()  # noqa: F405
        self.trunk_groups = TrunkGroups()  # noqa: F405
        self.users = Users()  # noqa: F405

        # authenticate newly instantiated object
        self._authenticate()

    def refresh_authorisation(self) -> bool:
        """Re-authenticates the session with the API. Can used if API key is due to expire.

        Raises:
            OSSessionRefreshFail: Raised if authentication fails.

        Returns:
            Bool: Returns True to indicate authentication was successful.
        """

        try:
            response = self.session.put_session()
            self._update_requester(response)
            return True
        except Exception:
            raise OSSessionRefreshFail()

    def get_auth_details(self):
        """Gets current session details.

        Raises:
            OSFailedToLocateSession: Raised when session details can't be found.
                Most likely because session has expired.

        Returns:
            Dict: Current session details.
        """

        try:
            return self.session.get_session()
        except Exception:
            raise OSFailedToLocateSession()

    def update_api(
        self,
        base_url: str = None,
        username: str = None,
        password: str = None,
        rate_limit: bool = None,
        logger: object = None,
    ):
        """Updates the API with new details.

        Args:
            base_url (str, optional): Base url of your odin instance api. Defaults to None.
            username (str, optional): Username used when logging into odin account. Defaults to None.
            password (str, optional): Password used when logging into odin account stored as virtual environment. Defaults to None.
            rate_limit (bool, optional): Enables (True) or Disables (False) rate limiting to 5 calls per second. Defaults to None.
        """

        if base_url:
            self.logger.info(
                f"API base_url updated, old: {self.base_url}, new: {base_url}"
            )
            self.base_url = base_url
            self._requester.base_url = base_url
        if username:
            self.logger.info(
                f"API username updated, old: {self.username}, new: {username}"
            )
            self.username = username
        if password:
            self.logger.info("API password updated")
            self._password = password
        if rate_limit:
            self.logger.info(
                f"API rate_limit updated, old: {self.rate_limit}, new: {rate_limit}"
            )
            self.rate_limit = rate_limit
            self._requester.rate_limit = rate_limit
        if logger:
            self.logger = logger
            self._requester.logger = logger
            self.logger.info("Logger updated")

    def _authenticate(self) -> bool:
        """Authenticates session with username and password supplied by user.

        Raises:
            OSApiAuthenticationFail: Raised if authenticaion fails.

        Returns:
            Bool: Returns True to indicate authentication was successful.
        """

        try:
            response = self.session.post_session(self.username, self._password)
            self._update_requester(response)
            return True
        except Exception:
            raise OSApiAuthenticationFail

    def _update_requester(self, session_response: dict):
        """When authenticating or re-auth update requester with token so it can make
        api calls

        Args:
            session_response (dict): Resposne from call.
        """

        self._requester.headers["Authorization"] = f"Bearer {session_response['token']}"
        self.authorised = True
        self.logger.info("API session updated with new token")

    def _setup_logger(self):
        logger = logging.getLogger("OS")
        logger.setLevel(logging.ERROR)
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "timestamp: %(asctime)s, level: %(levelname)s, module: %(module)s, function: %(funcName)s,  message: %(message)s"
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return logger

    def __str__(self) -> str:
        return (
            f"API - url: {self.base_url}, username: {self.username} "
            f"Authenticated: {self.authorised}"
        )



from .base_endpoint import BaseEndpoint


class CallProcessingPolicies(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    # POST

    # PUT

    def put_user_call_processing_policy(self, user_id: str, updates: dict):
        """
        Update the Call Processing Policies for a specified user.

        Args:
            user_id (str): The user ID of the user whose call processing policies need updating.
            updates (dict): Updates to apply to the specified user.

        Returns:
            Dict: Returns the updated call processing policies.
        """

        endpoint = "/users/call-processing-policy"

        updates["userId"] = user_id

        return self._requester.put(endpoint, data=updates)


# DELETE



import re

from ..exceptions import OSAliasNotFound


def locate_alias(alias, aliases: list):
    for a in aliases:
        if re.fullmatch(
            r"^[A-Za-z0-9\-_.!~*()']+$", alias
        ):  # As per Odin Spec: User alias cannot contain any characters except A-Z, a-z, 0-9, -_.!~*() or single quotes.
            return True


def main(api, service_provider_id: str, group_id: str, alias: str):
    RETRY_QUEUE = []
    MAX_RETRIES = 2
    OBJECT_WITH_ALIAS = []

    auto_attendants = api.auto_attendants.get_auto_attendants(
        service_provider_id, group_id
    )
    hunt_groups = api.hunt_groups.get_group_hunt_groups(service_provider_id, group_id)
    call_centers = api.call_centers.get_group_call_centers(
        service_provider_id, group_id
    )

    broadwork_entities_user_ids = []

    # save logger from api
    logger = api.logger

    for aa in auto_attendants:
        broadwork_entities_user_ids.append(["AA", aa["serviceUserId"]])

    for hg in hunt_groups:
        broadwork_entities_user_ids.append(["HG", hg["serviceUserId"]])

    for cc in call_centers:
        broadwork_entities_user_ids.append(["CC", cc["serviceUserId"]])

    logger.info("Fetching aa, hg, and cc")
    for broadwork_entity in broadwork_entities_user_ids:
        logger.info(f"Fetching '{broadwork_entity[1]}'")
        formatted = {
            "type": broadwork_entity[0],
            "service_user_id": broadwork_entity[1],
        }

        temp_object = ""

        try:
            if broadwork_entity[0] == "AA":
                temp_object = api.auto_attendants.get_auto_attendant(
                    broadwork_entity[1]
                )
            elif broadwork_entity[0] == "HG":
                temp_object = api.hunt_groups.get_group_hunt_group(broadwork_entity[1])
            else:
                temp_object = api.call_centers.get_group_call_center(
                    broadwork_entity[1]
                )

            formatted["name"] = temp_object["serviceInstanceProfile"]["name"]
            formatted["aliases"] = temp_object["serviceInstanceProfile"]["aliases"]

            OBJECT_WITH_ALIAS.append(formatted)

        except Exception:
            # add a retry count and add this entity to retry queue
            logger.error(
                f"Failed to fetch bre '{broadwork_entity[1]}' added to retry queue"
            )
            broadwork_entity.append(0)
            RETRY_QUEUE.append(broadwork_entity)

    # objects failed in first instance
    if RETRY_QUEUE:
        logger.info("Going through retry queue")
    while RETRY_QUEUE:
        entity_type, service_user_id, retry_count = RETRY_QUEUE.pop(
            0
        )  # Get the first item from the queue

        formatted = {}
        formatted["type"] = entity_type
        temp_object = ""

        try:
            if entity_type == "AA":
                temp_object = api.auto_attendants.get_auto_attendant(service_user_id)
            elif entity_type == "HG":
                temp_object = api.hunt_groups.get_group_hunt_group(service_user_id)
            else:
                temp_object = api.call_centers.get_group_call_center(service_user_id)

            formatted["name"] = temp_object["serviceInstanceProfile"]["name"]
            formatted["aliases"] = temp_object["serviceInstanceProfile"]["aliases"]

            OBJECT_WITH_ALIAS.append(formatted)

        except Exception:
            if retry_count < MAX_RETRIES:
                RETRY_QUEUE.append(
                    (entity_type, service_user_id, retry_count + 1)
                )  # Increment retry count and re-add to the queue
            else:
                logger.error(
                    f"Failed to process {entity_type} - {service_user_id} after {MAX_RETRIES} retries - skipping"
                )

    logger.info("Searching through aa, hg, and cc")
    for broadwork_entity in OBJECT_WITH_ALIAS:
        logger.info(f"Checking bre '{broadwork_entity['name']}'")
        if locate_alias(alias, broadwork_entity["aliases"]):
            logger.info(
                f"Alias found, type: user, service_user_id: {broadwork_entity['name']}, alias: {alias}"
            )
            return broadwork_entity
    logger.info(f"Alias '{alias}' not found in aa, hg, cc")

    logger.info("Fetching users")
    users = api.users.get_users(service_provider_id, group_id, extended=True)
    logger.info("Users successfully fetched")

    logger.info("Searching users")
    for user in users:
        logger.info(f"Checking {user['userId']}")
        if locate_alias(alias, user["aliases"]):
            logger.info(
                f"Alias found, type: user, user_id: {user['userId']}, alias: {alias}"
            )
            return {"type": "user", "user_id": user["userId"], "alias": alias}

    return OSAliasNotFound



from .base_endpoint import BaseEndpoint


class SharedCallAppearance(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    # POST

    def post_user_shared_call_appearance_endpoint(
        self, user_id: str, line_port: str, device_name: str
    ):
        """Creates a new Shared Call Apprance (SCA) on a single user.

        Args:
            user_id (str): Target user id of user to create SCA on.
            line_port (str): Line port to be assigned to the new SCA.
            device_name (_type_): Device to add for SCA from available devices.

        Returns:
            dict: New SCA details applied to user.
        """

        endpoint = "/users/shared-call-appearance/endpoints"

        data = {
            "userId": user_id,
            "linePort": line_port,
            "isActive": True,
            "allowOrigination": True,
            "allowTermination": True,
            "deviceName": device_name,
            "deviceLevel": "Group",
        }

        return self._requester.post(endpoint, data=data)

    # PUT

    # DELETE



from .base_endpoint import BaseEndpoint


class Reports(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_user_report(self, user_id: str):
        """Detailed report of user including services and service packs assigned.

        Args:
            user_id (str): Target user id of user.

        Returns:
            dict: Detailed report of user including services and service packs.
        """

        endpoint = "/users/reports/users"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)


# POST

# PUT

# DELETE



from .base_endpoint import BaseEndpoint


class Registration(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_user_registration(self, user_id: str):
        """Gets a users devices and if those devices are registered. This includes soft phones.

        Args:
            user_id (str): Target user ID to check registration

        Returns:
            dict: All users devices and details on device such as registration.
        """

        endpoint = "/users/registration"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)

    def get_bulk_user_registration(self, service_provider_id: str, group_id: str):
        """Gets all users in a group and their device registrations. This includes soft phones.

        Args:
            service_provider_id (str): Service Provider/ Enterprise ID where Group is hosted.
            group_id (str): Target Group ID where users are located.

        Returns:
            dict: All users devices and details on device such as registration.
        """

        endpoint = "/users/registration/bulk"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self._requester.get(endpoint, params=params)


# POST

# PUT

# DELETE



from typing import List
import json

from ..api import API
from ..utils import parsers
from . import broadwork_entities as bre


class DataStore:
    """Local store of objects, when each object is instantiated it is added to
    the appropriate list.

    Singleton design pattern to get store use get_instance().
    """

    def __init__(self):
        self.apis: List[API] = []
        self.service_providers_enterprises: List[bre.ServiceProvider] = []
        self.groups: List[bre.Group] = []
        self.trunk_groups: List[bre.TrunkGroup] = []
        self.auto_attendants: List[bre.AutoAttendant] = []
        self.call_centers: List[bre.CallCenter] = []
        self.hunt_groups: List[bre.HuntGroup] = []
        self.users: List[bre.User] = []
        self.devices: List[bre.Device] = []
        self.other_entities = []  # Non-common or custom objects

    def build_id_mapping(self):
        """
        Builds mapping of numbers IDs to entity.

        Example: {test@test.com: User, customID: CallCenter}
        """

        self.id_mapping = {}

        entities = (
            self.auto_attendants + self.call_centers + self.hunt_groups + self.users
        )

        for e in entities:
            try:
                self.id_mapping[e.id] = e
            except Exception:
                self.id_mapping[e.service_user_id] = e

    def build_number_mapping(self):
        """
        Builds mapping of numbers (phone numbers, extension, aliases) to entity.

        Example: {101: User, +1-123456789: CallCenter}
        """
        import re

        self.number_mapping = {}

        entities = (
            self.auto_attendants + self.call_centers + self.hunt_groups + self.users
        )

        for e in entities:
            if e.phone_number:
                self.number_mapping[e.phone_number] = e
            if e.extension:
                self.number_mapping[e.extension] = e
            if e.aliases:
                for a in e.aliases:
                    number = re.search(r"\d+", a).group()
                    self.number_mapping[number] = e

    def store_objects(self, *entities) -> None:
        """Takes in objects within the odin_api and custom and stores in lists
        depending on type.

        :param entity: broadwork entities used in odin_api
        """

        for e in entities:
            if isinstance(e, API):
                self.apis.append(e)
            elif isinstance(e, bre.ServiceProvider):
                self.service_providers_enterprises.append(e)
            elif isinstance(e, bre.Group):
                self.groups.append(e)
            elif isinstance(e, bre.TrunkGroup):
                self.trunk_groups.append(e)
            elif isinstance(e, bre.AutoAttendant):
                self.auto_attendants.append(e)
            elif isinstance(e, bre.CallCenter):
                self.call_centers.append(e)
            elif isinstance(e, bre.HuntGroup):
                self.hunt_groups.append(e)
            elif isinstance(e, bre.User):
                self.users.append(e)
            elif isinstance(e, bre.Device):
                self.devices.append(e)
            else:
                self.other_entities.append(e)

    def export_store(self) -> str:
        """Export all objects in the store and their relationships in JSON format."""
        export_data = {}
        object_lists = {
            "apis": self.apis,
            "service_providers": self.service_providers,
            "enterprises": self.enterprises,
            "groups": self.groups,
            "trunk_groups": self.trunk_groups,
            "auto_attendants": self.auto_attendants,
            "call_centers": self.call_centers,
            "hunt_groups": self.hunt_groups,
            "users": self.users,
            "devices": self.devices,
            "other_entities": self.other_entities,
        }

        for key, object_list in object_lists.items():
            export_data[key] = parsers.export_objects(object_list)

        return json.dumps(export_data, indent=2)

    def join_entities(entities):
        return "\n".join(str(entity) for entity in entities)

    def __str__(self) -> str:
        """returns complete list of entities in store."""
        entities = (
            self.apis
            + self.enterprises
            + self.service_providers
            + self.groups
            + self.trunk_groups
            + self.hunt_groups
            + self.users
        )

        return self.join_entities(entities)



from ...store import broadwork_entities as bre
from .report_entities import external_number

# nodes needed for graph generation
NODES = []
MAX_LEVEL = 30


def call_flow_module(node: object, data_store: object):
    # create mapping for easier searchability when gathering nodes
    data_store.build_number_mapping()

    # traverses the enities that forward to eachother and adds all to NODES
    _traverse_connecting_entities(node, data_store)

    return NODES


def _traverse_connecting_entities(entity: object, data_store: object, visited=None):
    if visited is None:
        visited = []

    if entity in visited:
        return
    else:
        visited.append(entity)

    try:
        if isinstance(entity, bre.User):
            # Checks to see if the number is external first and creates external number object for parsing later
            if (
                data_store.number_mapping.get(entity.call_forwarding_always) is None
                and len(str(entity.call_forwarding_always)) >= 3
                and str(entity.call_forwarding_always)[2].isdigit()
            ):
                entity.call_forwarding_always = external_number(
                    str(entity.call_forwarding_always)
                )
            else:
                entity.call_forwarding_always = data_store.number_mapping.get(
                    entity.call_forwarding_always
                )
                if (
                    entity.call_forwarding_always
                    and entity.call_forwarding_always not in visited
                ):
                    _traverse_connecting_entities(
                        entity.call_forwarding_always, data_store, visited
                    )

            if (
                data_store.number_mapping.get(entity.call_forwarding_busy) is None
                and len(str(entity.call_forwarding_busy)) >= 3
                and str(entity.call_forwarding_busy)[2].isdigit()
            ):
                entity.call_forwarding_busy = external_number(
                    str(entity.call_forwarding_busy)
                )
            else:
                entity.call_forwarding_busy = data_store.number_mapping.get(
                    entity.call_forwarding_busy
                )
                if (
                    entity.call_forwarding_busy
                    and entity.call_forwarding_busy not in visited
                ):
                    _traverse_connecting_entities(
                        entity.call_forwarding_busy, data_store, visited
                    )

            if (
                data_store.number_mapping.get(entity.call_forwarding_no_answer) is None
                and len(str(entity.call_forwarding_no_answer)) >= 3
                and str(entity.call_forwarding_no_answer)[2].isdigit()
            ):
                entity.call_forwarding_no_answer = external_number(
                    str(entity.call_forwarding_no_answer)
                )
            else:
                entity.call_forwarding_no_answer = data_store.number_mapping.get(
                    entity.call_forwarding_no_answer
                )
                if (
                    entity.call_forwarding_no_answer
                    and entity.call_forwarding_no_answer not in visited
                ):
                    _traverse_connecting_entities(
                        entity.call_forwarding_no_answer, data_store, visited
                    )

            if (
                data_store.number_mapping.get(entity.call_forwarding_not_reachable)
                is None
                and len(str(entity.call_forwarding_not_reachable)) >= 3
                and str(entity.call_forwarding_not_reachable)[2].isdigit()
            ):
                entity.call_forwarding_not_reachable = external_number(
                    str(entity.call_forwarding_not_reachable)
                )
            else:
                entity.call_forwarding_not_reachable = data_store.number_mapping.get(
                    entity.call_forwarding_not_reachable
                )
                if (
                    entity.call_forwarding_not_reachable
                    and entity.call_forwarding_not_reachable not in visited
                ):
                    _traverse_connecting_entities(
                        entity.call_forwarding_not_reachable, data_store, visited
                    )

        if isinstance(entity, bre.CallCenter):
            if entity.bounced_calls_enabled:
                # Checks to see if the number is external first and creates external number object for parsing later
                if (
                    data_store.number_mapping.get(
                        str(entity.bounced_calls_transfer_to_phone_number)
                    )
                    is None
                    and len(str(entity.bounced_calls_transfer_to_phone_number)) >= 3
                    and str(entity.bounced_calls_transfer_to_phone_number)[2].isdigit()
                ):
                    entity.bounced_calls_transfer_to_phone_number = external_number(
                        str(entity.bounced_calls_transfer_to_phone_number)
                    )
                else:
                    entity.bounced_calls_transfer_to_phone_number = (
                        data_store.number_mapping.get(
                            str(entity.bounced_calls_transfer_to_phone_number)
                        )
                    )
                    if (
                        entity.bounced_calls_transfer_to_phone_number
                        and entity.bounced_calls_transfer_to_phone_number not in visited
                    ):
                        _traverse_connecting_entities(
                            entity.bounced_calls_transfer_to_phone_number,
                            data_store,
                            visited,
                        )

            if entity.overflow_calls_action:
                if (
                    data_store.number_mapping.get(
                        str(entity.overflow_calls_transfer_to_phone_number)
                    )
                    is None
                    and len(str(entity.overflow_calls_transfer_to_phone_number)) >= 3
                    and str(entity.overflow_calls_transfer_to_phone_number)[2].isdigit()
                ):
                    entity.overflow_calls_transfer_to_phone_number = external_number(
                        str(entity.overflow_calls_transfer_to_phone_number)
                    )
                else:
                    entity.overflow_calls_transfer_to_phone_number = (
                        data_store.number_mapping.get(
                            str(entity.overflow_calls_transfer_to_phone_number)
                        )
                    )
                    if (
                        entity.overflow_calls_transfer_to_phone_number
                        and entity.overflow_calls_transfer_to_phone_number
                        not in visited
                    ):
                        _traverse_connecting_entities(
                            entity.overflow_calls_transfer_to_phone_number,
                            data_store,
                            visited,
                        )

            if entity.stranded_calls_action:
                if (
                    data_store.number_mapping.get(
                        str(entity.stranded_calls_transfer_to_phone_number)
                    )
                    is None
                    and len(str(entity.stranded_calls_transfer_to_phone_number)) >= 3
                    and str(entity.stranded_calls_transfer_to_phone_number)[2].isdigit()
                ):
                    entity.stranded_calls_transfer_to_phone_number = external_number(
                        str(entity.stranded_calls_transfer_to_phone_number)
                    )
                else:
                    entity.stranded_calls_transfer_to_phone_number = (
                        data_store.number_mapping.get(
                            str(entity.stranded_calls_transfer_to_phone_number)
                        )
                    )
                    if (
                        entity.stranded_calls_transfer_to_phone_number
                        and entity.stranded_calls_transfer_to_phone_number
                        not in visited
                    ):
                        _traverse_connecting_entities(
                            entity.stranded_calls_transfer_to_phone_number,
                            data_store,
                            visited,
                        )

            if entity.stranded_call_unavailable_action:
                if (
                    data_store.number_mapping.get(
                        str(entity.stranded_call_unavailable_transfer_to_phone_number)
                    )
                    is None
                    and len(
                        str(entity.stranded_call_unavailable_transfer_to_phone_number)
                    )
                    >= 3
                    and str(entity.stranded_call_unavailable_transfer_to_phone_number)[
                        2
                    ].isdigit()
                ):
                    entity.stranded_call_unavailable_transfer_to_phone_number = external_number(
                        str(entity.stranded_call_unavailable_transfer_to_phone_number)
                    )
                else:
                    entity.stranded_call_unavailable_transfer_to_phone_number = data_store.number_mapping.get(
                        str(entity.stranded_call_unavailable_transfer_to_phone_number)
                    )
                    if (
                        entity.stranded_call_unavailable_transfer_to_phone_number
                        and entity.stranded_call_unavailable_transfer_to_phone_number
                        not in visited
                    ):
                        _traverse_connecting_entities(
                            entity.stranded_call_unavailable_transfer_to_phone_number,
                            data_store,
                            visited,
                        )

        if isinstance(entity, bre.HuntGroup):
            if entity.forward_after_timeout_enabled:
                if (
                    data_store.number_mapping.get(
                        entity.no_answer_forward_to_phone_number
                    )
                    is None
                    and len(str(entity.no_answer_forward_to_phone_number)) >= 4
                    and str(entity.no_answer_forward_to_phone_number)[2].isdigit()
                ):
                    entity.no_answer_forward_to_phone_number = external_number(
                        str(entity.no_answer_forward_to_phone_number)
                    )
                else:
                    entity.no_answer_forward_to_phone_number = (
                        data_store.number_mapping.get(
                            str(entity.no_answer_forward_to_phone_number)
                        )
                    )
                    if (
                        entity.no_answer_forward_to_phone_number
                        and entity.no_answer_forward_to_phone_number not in visited
                    ):
                        _traverse_connecting_entities(
                            entity.no_answer_forward_to_phone_number,
                            data_store,
                            visited,
                        )

            if entity.call_forward_not_reachable_enabled:
                if (
                    data_store.number_mapping.get(
                        str(entity.call_forward_not_reachable_transfer_to_phone_number)
                    )
                    is None
                    and len(
                        str(entity.call_forward_not_reachable_transfer_to_phone_number)
                    )
                    >= 3
                    and str(entity.call_forward_not_reachable_transfer_to_phone_number)[
                        2
                    ].isdigit()
                ):
                    entity.call_forward_not_reachable_transfer_to_phone_number = external_number(
                        str(entity.call_forward_not_reachable_transfer_to_phone_number)
                    )
                else:
                    entity.call_forward_not_reachable_transfer_to_phone_number = data_store.number_mapping.get(
                        str(entity.call_forward_not_reachable_transfer_to_phone_number)
                    )
                    if (
                        entity.call_forward_not_reachable_transfer_to_phone_number
                        and entity.call_forward_not_reachable_transfer_to_phone_number
                        not in visited
                    ):
                        _traverse_connecting_entities(
                            entity.call_forward_not_reachable_transfer_to_phone_number,
                            data_store,
                            visited,
                        )

        if isinstance(entity, bre.AutoAttendant):
            for key in entity.business_hours_menu.keys:
                if "Transfer" in key.action:
                    if (
                        data_store.number_mapping.get(str(key.phone_number)) is None
                        and len(str(key.phone_number)) >= 3
                        and str(key.phone_number)[2].isdigit()
                    ):
                        key.phone_number = external_number(str(key.phone_number))
                    else:
                        key.phone_number = data_store.number_mapping.get(
                            str(key.phone_number)
                        )
                        if key.phone_number and key.phone_number not in visited:
                            _traverse_connecting_entities(
                                key.phone_number, data_store, visited
                            )

            for key in entity.after_hours_menu.keys:
                if "Transfer" in key.action:
                    if (
                        data_store.number_mapping.get(str(key.phone_number)) is None
                        and len(str(key.phone_number)) >= 3
                        and str(key.phone_number)[2].isdigit()
                    ):
                        key.phone_number = external_number(str(key.phone_number))
                    else:
                        key.phone_number = data_store.number_mapping.get(
                            str(key.phone_number)
                        )
                        if key.phone_number and key.phone_number not in visited:
                            _traverse_connecting_entities(
                                key.phone_number, data_store, visited
                            )
    except TypeError:
        # this object is further up the tree and a loop has occured - Move one and come back
        pass

    NODES.append(entity)  # Append outside of the if condition



from .base_endpoint import BaseEndpoint


class DNs(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_group_dns(self, service_provider_id: str, group_id: str):
        """Gets all numbers assigned to group.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where group is hosted.
            group_id (str): Group ID of target group.

        Returns:
            Dict: Dictionary containing all DNs assigned to group.
        """

        endpoint = "/groups/dns"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self._requester.get(endpoint, params=params)

    def get_group_dn_search(
        self,
        service_provider_id: str,
        group_id: str,
        dn: int,
        filter_type: str = None,
        limit: int = None,
    ):
        """Searches for numbers assigned to group and allows search criteria and limiting result.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where group is hosted.
            group_id (str): Group ID of target group where numbers are located.
            dn (int): Number/ part of number to search for.
            filter_type (str, optional): Options: equal to, starts with, or contains. Defaults to None.
            limit (int, optional): Limits the amount of values API returns. Defaults to None.

        Returns:
            List: List of numbers matching search criteria
        """
        from ..utils.formatters import format_filter_value

        endpoint = "/groups/dns/search"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        if filter_type:
            params["dn"] = format_filter_value(filter_type, dn)
        if limit:
            params["limit"] = limit

        return self._requester.get(endpoint, params=params)

    def get_group_dn_details(self, service_provider_id: str, group_id: str):
        """Gets all numbers assigned to Group in detail. This will show where the number is assigned
        in a group such as which user or hunt group.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where group is hosted.
            group_id (str): Group ID of target group where numbers are located.

        Returns:
            Dict: Dictionary of numbers with details of each number.
        """

        endpoint = "/groups/dns/details"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self._requester.get(endpoint, params=params)

    def get_system_dn_search(self, dn: int):
        """Searches for number from System level. This will return where the number is located on the system.
        It will show the Service Provider/ Enterprise, Group ID, and User ID the number is assigned to.

        Args:
            dn (int): Target number excluding country code e.g. 123456789

        Returns:
            List: List of dictionaries containing details of each number that fit the search criteria.
        """

        endpoint = "/system/dns/search"

        params = {"dn": f"+{dn}"}

        return self._requester.get(endpoint, params=params)

    def get_system_dn(self, dn: int):
        """Searches for number from System level. This will return where the number is located on the system.
        It will show the Service Provider/ Enterprise, Group ID, and User ID the number is assigned to.

        Args:
            dn (int): Target number excluding country code e.g. 123456789

        Returns:
            List: List of dictionaries containing details of each number that fit the search criteria.
        """

        endpoint = "/system/dns"

        params = {"phoneNumber": f"+{dn}"}

        return self._requester.get(endpoint, params=params)

    def get_system_dn_summary(self):
        """Returns all numbers assigned to system.

        Returns:
            List: List of all Service Providers/ Enterprises and numbers assigned in ranges.
        """

        endpoint = "/system/dns/summary"

        return self._requester.get(endpoint)

    def get_system_dn_utilization(self):
        """Returns DN statistics for each Service Provider/ Enterprise such as total DNs assigned.

        Returns:
            List: List of all Service Provider/ Enterprises with statistics of DNs for each.
        """

        endpoint = "/system/dns/utilization"

        return self._requester.get(endpoint)

    def get_service_provider_dn_search(
        self,
        service_provider_id: str,
        dn: int,
        filter_type: str = None,
        limit: int = None,
    ):
        """Searches for numbers assigned to Service Provider/ Enterprise and allows search criteria and limiting result.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where group is hosted.
            dn (int): Number/ part of number to search for.
            filter_type (str, optional): Options: equal to, starts with, or contains. Defaults to None.
            limit (int, optional): Limits the amount of values API returns. Defaults to None.

        Returns:
            List: List of numbers matching search criteria
        """
        from ..utils.formatters import format_filter_value

        endpoint = "/service-providers/dns/search"

        params = {"serviceProviderId": service_provider_id}

        if filter_type:
            params["dn"] = format_filter_value(filter_type, dn)
        if limit:
            params["limit"] = limit

        return self._requester.get(endpoint, params=params)

    def get_service_provider_dns(self, service_provider_id: str):
        """Returns all numbers assigned to Service Provider/ Enterprise with the group its assigned to
        and if the numbers can be deleted.

        Args:
            service_provider_id (str): Service Provider or Enterprise ID where target numbers are located.

        Returns:
            Dict: All numbers assigend to Service Provider/ Enterprise with group and delete status.
        """

        endpoint = "/service-providers/dns"

        params = {"serviceProviderId": service_provider_id}

        return self._requester.get(endpoint, params=params)

    # POST

    def post_group_dns(
        self,
        service_provider_id: str,
        group_id: str,
        start_of_range_number: str,
        end_of_range_number: str,
    ):
        """Adds a range of numbers to a Group. Range of numbers must be complete and
        format of number must follow: +{country code}-{number}.

        Adding a singular number - Set both the start and end of range parameters as the same number.

        Args:
            service_provider_id (str): Service provider ID where the target group is located.
            group_id (str): Group ID where numbers should be added to.
            start_of_range_number (str): Starting number in range to add to group.
            end_of_range_number (str): Ending number in range to add to group.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/dns"

        data = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "dns": [{"min": start_of_range_number, "max": end_of_range_number}],
        }

        return self._requester.post(endpoint, data=data)

    def post_group_dns_assign_bulk(
        self,
        service_provider_id: str,
        group_id: str,
        start_of_range_number: int,
        end_of_range_number: int,
    ):
        """Adds a range of numbers to a Group. Range of numbers must be complete and
        format of number must follow: +{country code}-{number}.

        Adding a singular number - Set both the start and end of range parameters as the same number.

        Args:
            service_provider_id (str): Service provider ID where the target group is located.
            group_id (str): Group ID where numbers should be added to.
            start_of_range_number (str): Starting number in range to add to group.
            end_of_range_number (str): Ending number in range to add to group.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/dns/assign/bulk"

        data = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "dns": [{"min": start_of_range_number, "max": end_of_range_number}],
        }

        return self._requester.post(endpoint, data=data)

    def post_group_dns_unassign_bulk(
        self,
        service_provider_id: str,
        group_id: str,
        start_of_range_number: int,
        end_of_range_number: int,
    ):
        """Unassign a range of numbers from a Group. Range of numbers must be complete and
        format of number must follow: +{country code}-{number}.

        Unassigning a singular number - Set both the start and end of range parameters as the same number.

        Args:
            service_provider_id (str): Service provider ID where the target group is located.
            group_id (str): Group ID where numbers where numbers are located.
            start_of_range_number (str): Starting number in range to unassign in group.
            end_of_range_number (str): Ending number in range to unassign in group.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/dns/unassign/bulk"

        data = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "dns": [{"min": start_of_range_number, "max": end_of_range_number}],
        }

        return self._requester.post(endpoint, data=data)

    def post_service_provider_dns(
        self,
        service_provider_id: str,
        start_of_range_number: int,
        end_of_range_number: int,
    ):
        """Adds a range of numbers to a Service Provider/ Enterprise. Range of numbers must be complete and
        format of number must follow: +{country code}-{number}.

        Adding a singular number - Set both the start and end of range parameters as the same number.

        Args:
            service_provider_id (str): Service provider ID where the target group is located.
            start_of_range_number (str): Starting number in range to add to group.
            end_of_range_number (str): Ending number in range to add to group.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/dns"

        data = {
            "serviceProviderId": service_provider_id,
            "dns": [{"min": start_of_range_number, "max": end_of_range_number}],
        }

        return self._requester.post(endpoint, data=data)

    # PUT

    def put_group_dns_activate(
        self, service_provider_id: str, group_id: str, activated: bool, numbers: list
    ):
        """Update activation state of a list of numbers assigned to a group.

        Args:
            service_provider_id (str): Service provider ID where the group is located.
            group_id (str): Group ID where numbers are hosted.
            activated (bool): True to activate number and False to deactivate.
            numbers (list): List of target numbers to update. These must be strings and follow correct format.

        Returns:
            JSON: All numbers assigned to group with activation state.
        """

        endpoint = "/groups/dns"

        data = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "dns": [
                {"activated": activated, "min": num, "max": None} for num in numbers
            ],
        }

        return self._requester.put(endpoint, data=data)

    # DELETE

    def delete_group_dns(
        self,
        service_provider_id: str,
        group_id: str,
        start_of_range_number: str,
        end_of_range_number: str,
    ):
        """Removes range of numbers from a Group.

        Args:
            service_provider_id (str): Service provider ID where the group is located.
            group_id (str): Group ID where numbers are hosted.
            start_of_range_number (str): Starting number in range to remove from group.
            end_of_range_number (str): Ending number in range to remove from group.

        Returns:
            None: This method does not return any specific value.
        """

        endpoint = "/groups/dns"

        data = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "dns": [
                {
                    "min": start_of_range_number,
                    "max": end_of_range_number,
                }
            ],
        }

        return self._requester.delete(endpoint, data=data)

    def service_provider_dns(
        self,
        service_provider_id: str,
        start_of_range_number: str,
        end_of_range_number: str,
    ):
        """Removes range of numbers from a Service Proiver or Enterprise.

        Args:
            service_provider_id (str): Service provider ID where target numbers are located.
            start_of_range_number (str): Starting number in range to remove from service provider.
            end_of_range_number (str): Ending number in range to remove from service provider.

        Returns:
            None: This method does not return any specific value.
        """
        endpoint = "/service-providers/dns"

        data = {
            "serviceProviderId": service_provider_id,
            "dns": [{"min": start_of_range_number, "max": end_of_range_number}],
        }

        return self._requester.delete(endpoint, data=data)



from .graphviz_module import GraphvizModule
from .parsing import call_flow_module
from .report_entities import call_flow, external_number, call_records_statistics

__all__ = [
    "GraphvizModule",
    "call_flow_module",
    "call_flow",
    "external_number",
    "call_records_statistics",
]



import os
import importlib
import inspect

# Get the directory path of this __init__.py file
current_dir = os.path.dirname(os.path.abspath(__file__))

# Iterate over files in the directory
for filename in os.listdir(current_dir):
    # Skip __init__.py, __pycache__, and non-Python files
    if (
        filename == "__init__.py"
        or filename.startswith("__")
        or not filename.endswith(".py")
        or filename == "report_utils"
    ):
        continue

    module_name = filename[:-3]  # remove the .py extension

    # Dynamically import the module
    module = importlib.import_module(f".{module_name}", package=__name__)

    # Check if there is a function named 'main' in the module
    if hasattr(module, "main") and inspect.isfunction(module.main):
        # Add it to the current package's namespace with the module_name as the attribute name
        globals()[module_name] = module.main



import os
import json
import copy

from pprint import pprint as pp


class ConfigManager:
    def __init__(self):
        """
        ConfigManager loads and caches template configs for popular Broadworks entities.

        This class gives the end user access to these templates so they have utilise them
        when making api call.

        Complete list of template configs (name formatted how user needs to reference):
        - auto_attendant
        - call_center
        - device
        - group
        - hunt_group
        - service_provider
        - trunk_group
        - user
        """
        self.config_dir = os.path.join(os.path.dirname(__file__), "configs")
        self._configs = {}
        self._load_configs()
        self.available_configs = [
            "auto_attendant",
            "call_center",
            "device",
            "group",
            "hunt_group",
            "service_provider",
            "trunk_group",
            "user",
        ]

    def _load_configs(self):
        """Load all JSON configuration files from the given directory."""
        for filename in os.listdir(self.config_dir):
            if filename.endswith(".json"):
                filepath = os.path.join(self.config_dir, filename)
                try:
                    with open(filepath, "r") as f:
                        config_data = json.load(f)
                    config_key = os.path.splitext(filename)[0]
                    self._configs[config_key] = config_data
                except Exception as e:
                    print(f"Error loading {filename}: {e}")

    def reload(self):
        """Clear the cache and reload all configuration files from disk."""
        self._configs = {}
        self._load_configs()

    def list_configs(self):
        """Return a list of available configuration names."""
        return self.available_configs

    def get_config(self, config_name: str):
        """Returns the full configuration of specified config type.

        Args:
            config_name (str, optional): Name of config to fetch.

        Raises:
            KeyError: Raised is config name is not found in supported configs.

        Returns:
            dict: Configuration specified.

        """

        if config_name not in self.available_configs:
            raise KeyError(f"Config '{config_name}' not found.")

        return self._configs[config_name]

    def get_value(self, config_name: str, key_path: str):
        """Gets specifec value from configuration. Dot notion is used to target
        the value needed.

        Args:
            config_name (str): Configuration with needed value.
            key_path (str): Dot notation path to value for example 'serviceProviderId.name'

        Raises:
            KeyError: Raised if config_name is not in supported configurations.
            KeyError: Raised if target value is not found in configuration.

        Returns:
            object: Value from configuarion.
        """

        if config_name not in self.available_configs:
            raise KeyError(f"Config '{config_name}' not found.")

        config = self.get_config(config_name)
        keys = key_path.split(".")
        value = config
        for k in keys:
            if isinstance(value, dict) and k in value:
                value = value[k]
            else:
                raise KeyError(
                    f"Key path '{key_path}' not found in config '{config_name}'."
                )
        return value

    def copy_config(self, config_name):
        """Creates a complete copy of configuration.

        Args:
            config_name (str, optional): Name of config to copy.

        Raises:
            KeyError: Config name not found.

        Returns:
            dict: Configuration of specified config requested.

        """
        if config_name not in self.available_configs:
            KeyError(f"Config '{config_name}' not found.")

        if config_name in self._configs:
            return copy.deepcopy(self._configs[config_name])

    def view_config(self, config_name: str):
        """Prints template configuration in a human friendly way to digest it.

        Args:
            config_name (str): Name of supported configuration to view.

        Raises:
            KeyError: Raised of config_name is not found in supported configurations.

        Returns:
            str: Formatted configuration.
        """

        if config_name not in self.available_configs:
            KeyError(f"Config '{config_name}' not found.")

        return pp(self._configs[config_name])



from .base_endpoint import BaseEndpoint


class AlternateNumbers(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_user_alternate_numbers(self, user_id: str):
        """Fetches a list of a user/ service such as Auto Attendant, Hunt Group, or Call Centres
        alternate numebrs.

        Args:
            user_id (str): Target user/ service_user_id

        Returns:
            Dict: List of all alternate numbers assigned to the user/ service.
        """

        endpoint = "/users/alternate-numbers"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)

    # POST

    # PUT

    # DELETE



from .base_endpoint import BaseEndpoint


class Schedules(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_group_schedules(self, service_provider_id: str, group_id: str):
        """Retrieves the Business Schedules for the specified group.

        Args:
            service_provider_id (str): Target Service Provider ID where group is hosted.
            group_id (str): Target Group ID with schedules.

        Returns:
            List: List of all the groups schedules, including Name, Type and Level.
        """

        endpoint = "/groups/schedules"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self._requester.get(endpoint, params=params)

    def get_group_events(
        self, service_provider_id: str, group_id: str, name: str, type: str
    ):
        """Retrieves the Business Schedule's Events for the specified group.

        Args:
            service_provider_id (str): Target Service Provider ID
            group_id (str): Target Group ID
            name (str): Name of the target Busisness Schedule
            type (str): The type of the Business Schedule (Time, Holiday)

        Returns:
            List: List of each Business Schedule Event, including startTime and endTime.
        """

        endpoint = "/groups/events"

        params = {
            "serviceProviderId": service_provider_id,
            "groupId": group_id,
            "name": name,
            "type": type,
        }

        return self._requester.get(endpoint, params=params)


# POST

# PUT

# DELETE



"""Library exceptions."""


class OSError(Exception):
    """Odin's Spears Exceptions"""

    def __str__(self) -> str:
        return "I dont think you can be trusted in a combat situation"


# API


class OSApiResponseError(OSError):
    """Raised when Odin Api returns an error code."""

    def __init__(self, response):
        response = response.json()
        self.response = (
            f"{response['details']} {response['status']}: {response['error']}"
        )

    def __str__(self) -> str:
        return self.response


class OSRequestTypeError(OSError):
    """Raised when unsupport request type is given."""

    def __str__(self) -> str:
        return "Non-supported request type, supported: GET, POST, PUT, DELETE."


# AUTHENTICATION


class OSApiAuthenticationFail(OSError):
    """Raised when api fails to authenticate"""

    def __str__(self) -> str:
        return "Failed to authenticate. Check username, password, and url."


class OSSessionRefreshFail(OSError):
    """Raised when refreshing session fails."""

    def __str__(self) -> str:
        return (
            "Refreshing sesion failed. Check credentials are valid and "
            "token has not yet expired. If expired request another."
        )


class OSLogoutFailed(OSError):
    """Raised when logout attempt failed."""

    def __str__(self) -> str:
        return "Failed to logout, session still valid. Please try again."


class OSFailedToLocateSession(OSError):
    """Raised when user attempts to get session details but session cant be found."""

    def __str__(self) -> str:
        return "Session details not found. Check token is valid and not exppired."


# NOT-FOUND


class OSExtensionNotFound(OSError):
    """Raised when a searched extension is not found"""

    def __str__(self) -> str:
        return "Cannot locate extension. Please alter search criteria"


# FILTER


class OSUnsupportedFilter(OSError):
    """Raised when user requests to filter on unsupported filter"""

    def __init__(self, filter_attempt):
        self.filter_attempt = filter_attempt

    def __str__(self) -> str:
        return f"Filter '{self.filter_attempt}' is unsupported. Supported: macAddress, lastName, firstName, dn, emailAddress, userId, extension"


class OSUnsupportedFilterType(OSError):
    """Raised when user requests to filter on unsupported filter"""

    def __init__(self, type_attempt):
        self.type_attempt = type_attempt

    def __str__(self) -> str:
        return f"Filter type '{self.type_attempt}' is unsupported. Supported: contains, startsWith, endsWith, equals"


# FILES


class OSFileNotFound(OSError):
    """Raised when a file can not be found."""

    def __str__(self) -> str:
        return "File can not be found, please check path and file name."


# FORMATTING


class OSInvalidCode(OSError):
    """Raised when code is less than 4 and higher than 6."""

    def __str__(self) -> str:
        return "Code needs to be between 4 and 6 digits."


class OSInvalidWeighting(OSError):
    """Raised when invalid weighted call distribution set."""

    def __str__(self) -> str:
        return "Weighted call distribution is not equal to 100. This weight must add up to 100."


class OSInvalidData(OSError):
    """Raised when data used in request is invalid or incomplete."""

    def __str__(self) -> str:
        return (
            "Data invalid or incomplete, please check data passed to method is correct."
        )


class OSInvalidBroadworkService(OSError):
    """Raised when service given by user is not a valid license of Broadworks."""

    def __str__(self) -> str:
        return "Service invalid, please check the services you have given are valid Broadwork services."


class OSInvalidPasswordType(OSError):
    """Raised when password requested is invalid or not supported."""

    def __str__(self) -> str:
        return "Invalid or unsupported password, please review supported passwords."


#  OS FEATURES


class OSAliasNotFound(OSError):
    """Raised when alias is not found in Broadowks Group."""

    def __str__(self) -> str:
        return "Alias not found, it either does not exist or check alias."


class OSServiceNotAssigned(OSError):
    """Raised a service needed is not assigned to a Broadworks entity."""

    def __str__(self) -> str:
        return "Service not assigend to target Broadworks entity. Please check services assigned."


# RANGE


class OSRangeFault(OSError):
    """Raised when a numeric range is invalid to the context provided,
    E.G: uninitialised range values, disordered range values.
    """

    def __str__(self):
        return "Range fault raised. Please verify integrity of passed range values"


# PARSING


class OSObjectParseError(OSError):
    """Raised when parsing Broadworks Entity fails."""

    def __str__(self) -> str:
        return "Parsing Broadwork Entity failed."



from .base_endpoint import BaseEndpoint


class CallRecords(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET
    def get_users_stats(
        self,
        user_id: str,
        start_date: str,
        end_date: str = None,
        start_time: str = "00:00:00",
        end_time: str = "23:59:59",
        time_zone: str = "Z",
    ):
        """Pulls a single users call statistics for a specified period of time. 

        Args:
            user_id (str): Target user ID you would like to pull call statistics for.
            start_date (str): Start date of desired time period. Date must follow format 'YYYY-MM-DD'
            end_date (str, optional): End date of desired time period. Date must follow format 'YYYY-MM-DD'.\
                If this date is the same as Start date you do not need this parameter. Defaults to None.
            start_time (_type_, optional): Start time of desired time period. Time must follow formate 'HH:MM:SS'. \
                If you do not need to filter by time and want the whole day leave this parameter. Defaults to "00:00:00". MAX Request is 3 months.
            end_time (_type_, optional): End time of desired time period. Time must follow formate 'HH:MM:SS'. \
                If you do not need to filter by time and want the whole day leave this parameter. Defaults to "23:59:59". MAX Request is 3 months.
            time_zone (str, optional): A specified time you would like to see call records in. \
                Time zone must follow format 'GMT', 'EST', 'PST'. Defaults to "Z" (UTC Time Zone).

        Returns:
            Dict: Users call record statistics for specified time period.
        """

        # checks if end_date has been left and therefore we assume user wants same date.
        if not end_date:
            end_date = start_date

        endpoint = "/users/call-records/stats"

        params = {
            "userIds": user_id,
            "startTime": f"{start_date}T{start_time}{time_zone}",
            "endTime": f"{end_date}T{end_time}{time_zone}",
        }

        return self._requester.get(endpoint, params=params)

    # POST
    # PUT
    # DELETE



from .base_endpoint import BaseEndpoint


class Services(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_user_services_assigned(self, user_id: str):
        """

        Args:
            user_id (str): _description_
        """
        endpoint = "/users/services/assigned"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)

    def get_user_services(self, user_id: str):
        """Fetch all services assigned to a broadwrok entity, this can be
        a user, AA, CC, or HG.

        Args:
            user_id (str): User ID of the target Broadworks entity.

        Returns:
            Dict: Broadwork entiy and a list of services assigned.
        """

        endpoint = "/users/services"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)

    def get_user_service_settings(self, user_id: str):
        """Retrieves all service settings for a specific user.

        Args:
            user_id (str): ID of the target user

        Returns:
            Dict: A dictionary containing all the service settings for the specified user.
        """

        endpoint = "/users/services/settings"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)

    def get_group_services(self, group_id: str, service_provider_id: str):
        """
        Fetch all userServices, groupServices and servicePackServices assigned to a group.

        Args:
            group_id (str): GroupID of the target
            service_provider_id (str): Service Provider or Enterprise ID of the target.

        Returns:
            Dict: Authorised and assigned services within the group.
        """

        endpoint = "/groups/services"

        params = {"groupId": group_id, "serviceProviderId": service_provider_id}

        return self._requester.get(endpoint, params=params)

    def get_group_services_user_assigned(
        self,
        group_id: str,
        service_provider_id: str,
        service_name: str,
        service_type: str,
    ):
        """
        Get details of the user/service instances where a particular service is assigned.

        Args:
            group_id (str): GroupID being targeted
            service_provider_id (str): Service provider/Enterprise ID where the group is located.
            service_name (str): Name of the service querying
            service_type (str): Type of service. Either: serviceName or servicePackName

        Returns:
            Dict: Users/Service Instances where the service is assigned.
        """

        endpoint = "/groups/services/assigned"

        params = {
            "groupId": group_id,
            "serviceProviderId": service_provider_id,
            "serviceType": service_type,
            "serviceName": service_name,
        }

        return self._requester.get(endpoint, params=params)

    # POST

    # PUT

    def put_user_services(
        self,
        user_id: str,
        services: list = None,
        service_packs: list = None,
        assigned: bool = True,
    ):
        """Update the services assigned to a user. NOT service/feature packs.

        Args:
            user_id (str): User ID of the target user.
            services (list): List of services to be applied to user.
            service_packs (list): List of service packs to be applied to user.
            assigned (bool, optional): Assign (True) or unassign(False). Defaults to True.

        Returns:
            Dict: User services assigned to the user.
        """

        endpoint = "/users/services"

        data = {"userId": user_id}

        if services:
            data["userServices"] = [
                {"serviceName": service, "assigned": assigned} for service in services
            ]
        if service_packs:
            data["servicePackServices"] = [
                {"serviceName": service_pack, "assigned": assigned}
                for service_pack in service_packs
            ]
        return self._requester.put(endpoint, data=data)

    def put_user_service_settings(self, user_id: str, settings: dict):
        """Updates specific service settings for a given user.
        This function allows you to modify one or more service settings associated with a particular user.

        Args:
            user_id (str): The ID of the target user
            settings (dict): A dictionary containing the new settings to be applied. The structure of this dictionary should mirror the API's expected format for updating service settings.

        Returns:
            Dict: A dictionary representing the updated service settings for the specified user.
        """

        endpoint = "/users/services/settings"

        data = {"userId": user_id, **settings}

        return self._requester.put(endpoint, data=data)


# DELETE



from .base_endpoint import BaseEndpoint


class CallForwardingAlways(BaseEndpoint):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # GET

    def get_user_call_forwarding_always(self, user_id: str):
        """Retrieves the Forwarding Always status for the specified User.

        Args:
            user_id (str): Target User ID

        Returns:
            Dict: Forwarding enabled status, and the Number to be Forwarded to.
        """

        endpoint = "/users/call-forwarding-always"

        params = {"userId": user_id}

        return self._requester.get(endpoint, params=params)

    def get_bulk_call_forwarding_always(self, service_provider_id: str, group_id: str):
        """Retrieves the Forwarding Always status for all users within a specified group.

        Args:
            service_provider_id (str): Target Service Provider where group is hosted
            group_id (str): Target Group ID

        Returns:
            List: Forwarding enabled status, the Number to be Forwarded to, and User information.
        """

        endpoint = "/users/call-forwarding-always/bulk"

        params = {"serviceProviderId": service_provider_id, "groupId": group_id}

        return self._requester.get(endpoint, params=params)


# POST

# PUT

# DELETE
